@@ -1,4 +1,4 @@
-use crate::types::{Asn1Writable, SimpleAsn1Writable};
+use crate::types::Asn1Writable;
 use crate::Tag;
 #[cfg(not(feature = "std"))]
 use alloc::vec::Vec;
@@ -46,21 +46,26 @@ impl WriteBuf {
         self.0.as_mut_slice()
     }
 
+    // Reserve space for up to `len` additional bytes.
     #[inline]
-    pub fn push_byte(&mut self, b: u8) -> WriteResult {
+    pub fn reserve_additional(&mut self, len: usize) -> WriteResult {
         self.0
-            .try_reserve(1)
+            .try_reserve(len)
             .map_err(|_| WriteError::AllocationError)?;
 
+        Ok(())
+    }
+
+    #[inline]
+    pub fn push_byte(&mut self, b: u8) -> WriteResult {
+        self.reserve_additional(1)?;
         self.0.push(b);
         Ok(())
     }
 
     #[inline]
     pub fn push_slice(&mut self, data: &[u8]) -> WriteResult {
-        self.0
-            .try_reserve(data.len())
-            .map_err(|_| WriteError::AllocationError)?;
+        self.reserve_additional(data.len())?;
 
         self.0.extend_from_slice(data);
         Ok(())
@@ -77,6 +82,16 @@ fn _length_length(length: usize) -> u8 {
     num_bytes
 }
 
+/// Calculate the number of bytes needed to encode a length field for the given content length.
+/// This includes the length-of-length byte for lengths >= 128.
+pub(crate) fn length_encoding_size(content_length: usize) -> usize {
+    if content_length < 128 {
+        1
+    } else {
+        1 + _length_length(content_length) as usize
+    }
+}
+
 fn _insert_at_position(buf: &mut WriteBuf, pos: usize, data: &[u8]) -> WriteResult {
     for _ in 0..data.len() {
         buf.push_byte(0)?;
@@ -104,69 +119,53 @@ impl Writer<'_> {
     /// Writes a single element to the output.
     #[inline]
     pub fn write_element<T: Asn1Writable>(&mut self, val: &T) -> WriteResult {
+        if let Some(len) = val.encoded_length() {
+            self.buf.reserve_additional(len)?;
+        }
         val.write(self)
     }
 
-    /// This is an alias for `write_element::<Explicit<T, tag>>`.
-    pub fn write_explicit_element<T: Asn1Writable>(&mut self, val: &T, tag: u32) -> WriteResult {
-        let tag = crate::explicit_tag(tag);
-        self.write_tlv(tag, |dest| Writer::new(dest).write_element(val))
-    }
-
-    /// This is an alias for `write_element::<Option<Explicit<T, tag>>>`.
-    pub fn write_optional_explicit_element<T: Asn1Writable>(
-        &mut self,
-        val: &Option<T>,
-        tag: u32,
-    ) -> WriteResult {
-        if let Some(v) = val {
-            let tag = crate::explicit_tag(tag);
-            self.write_tlv(tag, |dest| Writer::new(dest).write_element(v))
-        } else {
-            Ok(())
-        }
-    }
-
-    /// This is an alias for `write_element::<Implicit<T, tag>>`.
-    pub fn write_implicit_element<T: SimpleAsn1Writable>(
-        &mut self,
-        val: &T,
-        tag: u32,
-    ) -> WriteResult {
-        let tag = crate::implicit_tag(tag, T::TAG);
-        self.write_tlv(tag, |dest| val.write_data(dest))
-    }
-
-    /// This is an alias for `write_element::<Option<Implicit<T, tag>>>`.
-    pub fn write_optional_implicit_element<T: SimpleAsn1Writable>(
-        &mut self,
-        val: &Option<T>,
-        tag: u32,
-    ) -> WriteResult {
-        if let Some(v) = val {
-            let tag = crate::implicit_tag(tag, T::TAG);
-            self.write_tlv(tag, |dest| v.write_data(dest))
-        } else {
-            Ok(())
-        }
-    }
-
     /// Writes a TLV with the specified tag where the value is any bytes
     /// written to the `Vec` in the callback. The length portion of the
     /// TLV is automatically computed.
+    ///
+    /// If `content_length` is provided, it reduces the number of
+    /// re-allocations required.
     #[inline]
     pub fn write_tlv<F: FnOnce(&mut WriteBuf) -> WriteResult>(
         &mut self,
         tag: Tag,
+        content_length: Option<usize>,
         body: F,
     ) -> WriteResult {
         tag.write_bytes(self.buf)?;
-        // Push a 0-byte placeholder for the length. Needing only a single byte
-        // for the element is probably the most common case.
-        self.buf.push_byte(0)?;
-        let start_len = self.buf.len();
-        body(self.buf)?;
-        self.insert_length(start_len)
+
+        match content_length {
+            Some(len) => {
+                // Optimized path: write the correct length encoding upfront
+                if len < 128 {
+                    self.buf.push_byte(len as u8)?;
+                } else {
+                    let num_length_bytes = _length_length(len);
+                    self.buf.push_byte(0x80 | num_length_bytes)?;
+                    for i in (1..=num_length_bytes).rev() {
+                        self.buf.push_byte((len >> ((i - 1) * 8)) as u8)?;
+                    }
+                }
+                let start_len = self.buf.len();
+                body(self.buf)?;
+                assert_eq!(len, self.buf.len() - start_len);
+                Ok(())
+            }
+            None => {
+                // Write a placeholder and then fix the length up later as
+                // required.
+                self.buf.push_byte(0)?;
+                let start_len = self.buf.len();
+                body(self.buf)?;
+                self.insert_length(start_len)
+            }
+        }
     }
 
     #[inline]
@@ -219,7 +218,7 @@ mod tests {
         Enumerated, Explicit, GeneralizedTime, IA5String, Implicit, ObjectIdentifier,
         OctetStringEncoded, OwnedBigInt, OwnedBigUint, OwnedBitString, PrintableString, Sequence,
         SequenceOf, SequenceOfWriter, SequenceWriter, SetOf, SetOfWriter, Tlv, UniversalString,
-        UtcTime, Utf8String, VisibleString, WriteError,
+        UtcTime, Utf8String, VisibleString, WriteError, X509GeneralizedTime,
     };
     #[cfg(not(feature = "std"))]
     use alloc::vec::Vec;
@@ -553,23 +552,59 @@ mod tests {
     }
 
     #[test]
-    fn test_write_generalizedtime() {
+    fn test_write_x509_generalizedtime() {
         assert_writes(&[
             (
-                GeneralizedTime::new(DateTime::new(1991, 5, 6, 23, 45, 40).unwrap()).unwrap(),
+                X509GeneralizedTime::new(DateTime::new(1991, 5, 6, 23, 45, 40).unwrap()).unwrap(),
                 b"\x18\x0f19910506234540Z",
             ),
             (
-                GeneralizedTime::new(DateTime::new(1970, 1, 1, 0, 0, 0).unwrap()).unwrap(),
+                X509GeneralizedTime::new(DateTime::new(1970, 1, 1, 0, 0, 0).unwrap()).unwrap(),
                 b"\x18\x0f19700101000000Z",
             ),
             (
-                GeneralizedTime::new(DateTime::new(2009, 11, 15, 22, 56, 16).unwrap()).unwrap(),
+                X509GeneralizedTime::new(DateTime::new(2009, 11, 15, 22, 56, 16).unwrap()).unwrap(),
                 b"\x18\x0f20091115225616Z",
             ),
         ]);
     }
 
+    #[test]
+    fn test_write_generalizedtime() {
+        assert_writes(&[
+            (
+                GeneralizedTime::new(DateTime::new(1991, 5, 6, 23, 45, 40).unwrap(), Some(1_234))
+                    .unwrap(),
+                b"\x18\x1919910506234540.000001234Z",
+            ),
+            (
+                GeneralizedTime::new(DateTime::new(1991, 5, 6, 23, 45, 40).unwrap(), Some(1))
+                    .unwrap(),
+                b"\x18\x1919910506234540.000000001Z",
+            ),
+            (
+                GeneralizedTime::new(DateTime::new(1970, 1, 1, 0, 0, 0).unwrap(), None).unwrap(),
+                b"\x18\x0f19700101000000Z",
+            ),
+            (
+                GeneralizedTime::new(
+                    DateTime::new(2009, 11, 15, 22, 56, 16).unwrap(),
+                    Some(100_000_000),
+                )
+                .unwrap(),
+                b"\x18\x1120091115225616.1Z",
+            ),
+            (
+                GeneralizedTime::new(
+                    DateTime::new(2009, 11, 15, 22, 56, 16).unwrap(),
+                    Some(999_999_999),
+                )
+                .unwrap(),
+                b"\x18\x1920091115225616.999999999Z",
+            ),
+        ]);
+    }
+
     #[test]
     fn test_write_enumerated() {
         assert_writes::<Enumerated>(&[
@@ -598,6 +633,14 @@ mod tests {
             .unwrap(),
             b"\x30\x03\x01\x01\xff"
         );
+        assert_eq!(
+            write(|w| {
+                w.write_element(&SequenceWriter::new(&|w: &mut Writer<'_>| {
+                    w.write_element(&b"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa")
+                }))
+            }).unwrap(),
+            b"\x30\x81\x84\x04\x81\x81aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
+        );
 
         assert_writes(&[(
             parse_single::<Sequence<'_>>(b"\x30\x06\x01\x01\xff\x02\x01\x06").unwrap(),
@@ -680,39 +723,6 @@ mod tests {
             (Implicit::new(true), b"\x82\x01\xff"),
             (Implicit::new(false), b"\x82\x01\x00"),
         ]);
-
-        assert_eq!(
-            write(|w| { w.write_optional_implicit_element(&Some(true), 2) }).unwrap(),
-            b"\x82\x01\xff"
-        );
-        assert_eq!(
-            write(|w| { w.write_optional_explicit_element::<u8>(&None, 2) }).unwrap(),
-            b""
-        );
-
-        assert_eq!(
-            write(|w| {
-                w.write_optional_implicit_element(&Some(SequenceWriter::new(&|_w| Ok(()))), 2)
-            })
-            .unwrap(),
-            b"\xa2\x00"
-        );
-        assert_eq!(
-            write(|w| { w.write_optional_explicit_element::<SequenceWriter<'_>>(&None, 2) })
-                .unwrap(),
-            b""
-        );
-
-        assert_eq!(
-            write(|w| { w.write_implicit_element(&true, 2) }).unwrap(),
-            b"\x82\x01\xff"
-        );
-
-        assert_eq!(
-            write(|w| { w.write_implicit_element(&SequenceWriter::new(&|_w| { Ok(()) }), 2) })
-                .unwrap(),
-            b"\xa2\x00"
-        );
     }
 
     #[test]
@@ -721,20 +731,6 @@ mod tests {
             (Explicit::new(true), b"\xa2\x03\x01\x01\xff"),
             (Explicit::new(false), b"\xa2\x03\x01\x01\x00"),
         ]);
-
-        assert_eq!(
-            write(|w| { w.write_optional_explicit_element(&Some(true), 2) }).unwrap(),
-            b"\xa2\x03\x01\x01\xff"
-        );
-        assert_eq!(
-            write(|w| { w.write_optional_explicit_element::<u8>(&None, 2) }).unwrap(),
-            b""
-        );
-
-        assert_eq!(
-            write(|w| { w.write_explicit_element(&true, 2) }).unwrap(),
-            b"\xa2\x03\x01\x01\xff"
-        );
     }
 
     #[test]
